/*
 * manager-routes.c
 * This file is part of Network-inador
 *
 * Copyright (C) 2025 - Félix Arreola Rodríguez
 *
 * Network-inador is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Network-inador is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Network-inador; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */

#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <linux/if.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "manager-private.h"
#include "network-inador-public.h"
#include "network-inador-manager.h"

/* Empaquetador de datos */
int _manager_pack_route_info (unsigned char *buffer, Route *route) {
	uint8_t route_family, route_type, route_prefix;
	uint8_t route_protocol, route_tos, route_scope;
	uint8_t has_prefsrc, has_gw;
	uint32_t route_table, route_priority, out_index;
	int family_size = 0;
	int num_nexthops, g, pos;
	union {
		struct in_addr v4;
		struct in6_addr v6;
	} dest, prefsrc, gw;
	RouteNH *nexthop;
	
	route_family = network_inador_route_get_family (route);
	
	if (route_family == AF_INET) {
		family_size = sizeof (struct in_addr);
	} else if (route_family == AF_INET6) {
		family_size = sizeof (struct in6_addr);
	}
	
	route_type = network_inador_route_get_type (route);
	route_table = network_inador_route_get_table (route);
	route_prefix = network_inador_route_get_prefix (route);
	route_protocol = network_inador_route_get_protocol (route);
	route_tos = network_inador_route_get_tos (route);
	route_scope = network_inador_route_get_scope (route);
	has_prefsrc = network_inador_route_has_prefsrc (route);
	route_priority = network_inador_route_get_priority (route);
	
	buffer[0] = route_family;
	buffer[1] = route_type;
	memcpy (&buffer[2], &route_table, 4);
	buffer[6] = route_prefix;
	buffer[7] = route_protocol;
	buffer[8] = route_tos;
	buffer[9] = route_scope;
	buffer[10] = 0;
	if (has_prefsrc) {
		/* Tiene pref-src */
		buffer[10] |= 0x01;
	}
	
	num_nexthops = network_inador_route_get_num_nexthops (route);
	
	buffer[11] = num_nexthops;
	memcpy (&buffer[12], &route_priority, 4);
	
	/* Copiar la Ruta de destino */
	network_inador_route_get_dest (route, &dest, NULL);
	memcpy (&buffer[16], &dest, family_size);
	pos = 16 + family_size;
	
	/* Copiar el pref-src, si lo tiene */
	if (has_prefsrc) {
		network_inador_route_get_prefsrc (route, &prefsrc, NULL);
		memcpy (&buffer[pos], &prefsrc, family_size);
		pos += family_size;
	}
	
	/* TODO: Recorrer todos los nexthops */
	for (g = 0; g < num_nexthops; g++) {
		nexthop = network_inador_route_get_nexthop_by_index (route, g);
		
		buffer[pos] = 0;
		has_gw = network_inador_nexthop_has_gw (nexthop);
		if (has_gw) {
			buffer[pos] |= 0x01;
		}
		pos++;
		buffer[pos++] = network_inador_nexthop_get_flags (nexthop); /* Flags next-hop */
		buffer[pos++] = network_inador_nexthop_get_weight (nexthop); /* Weight next-hop */
		
		buffer[pos++] = 0; /* Omitimos un byte faltante */
		
		/* La interfaz de salida */
		out_index = network_inador_nexthop_get_out_index (nexthop);
		memcpy (&buffer[pos], &out_index, 4);
		pos += 4;
		
		if (has_gw) {
			network_inador_nexthop_get_gw (nexthop, &gw, NULL);
			memcpy (&buffer[pos], &gw, family_size);
			pos += family_size;
		}
	}
	
	return pos;
}

/* Los comandos que verdaderamente envian */
void _manager_send_route (ManagerClientInfo *manager_client, Route *route, uint8_t is_event) {
	unsigned char buffer[100];
	int size;
	
	if (is_event) {
		buffer[0] = NET_INADOR_TYPE_EVENT;
		buffer[1] = NET_INADOR_EVENT_ROUTE_ADDED;
	} else {
		buffer[0] = NET_INADOR_TYPE_RESPONSE;
		buffer[1] = NET_INADOR_RESPONSE_ROUTE;
	}
	
	size = 2 + _manager_pack_route_info (&buffer[2], route);
	
	send (manager_client->fd, buffer, size, 0);
}

void _manager_route_handle_list_routes (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	int family;
	Route *route;
	uint32_t wanted_table, route_table;
	int g, route_count;
	
	if (buffer_len < 7) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_LIST_ROUTES);
		return;
	}
	
	family = buffer[6];
	
	if (family != AF_UNSPEC && family != AF_INET && family != AF_INET6) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_FAMILY, NET_INADOR_COMMAND_LIST_ROUTES);
		
		return;
	}
	
	memcpy (&wanted_table, &buffer[2], 4);
	
	if (family == AF_UNSPEC || family == AF_INET) {
		route_count = network_inador_get_routes_count (manager_client->manager->handle, AF_INET);
		for (g = 0; g < route_count; g++) {
			route = network_inador_get_route_by_index (manager_client->manager->handle, AF_INET, g);
			
			route_table = network_inador_route_get_table (route);
			if (wanted_table != 0 && route_table != wanted_table) continue;
			
			_manager_send_route (manager_client, route, FALSE);
		}
	}
	
	if (family == AF_UNSPEC || family == AF_INET6) {
		route_count = network_inador_get_routes_count (manager_client->manager->handle, AF_INET6);
		for (g = 0; g < route_count; g++) {
			route = network_inador_get_route_by_index (manager_client->manager->handle, AF_INET6, g);
			
			route_table = network_inador_route_get_table (route);
			if (wanted_table != 0 && route_table != wanted_table) continue;
			
			_manager_send_route (manager_client, route, FALSE);
		}
	}
	
	_manager_send_end_command (manager_client, NET_INADOR_COMMAND_LIST_ROUTES);
}

/* Los eventos que vienen desde la librería */
void _manager_route_added_event_cb (NetworkInadorHandle *handle, Route *route, void *data) {
	Manager *manager = (Manager *) data;
	GList *g;
	ManagerClientInfo *manager_client;
	uint8_t family, prefix;
	union {
		struct in_addr v4;
		struct in6_addr v6;
	} dest;
	uint32_t table;
	char buffer_ip[200];
	
	family = network_inador_route_get_family (route);
	table = network_inador_route_get_table (route);
	prefix = network_inador_route_get_prefix (route);
	network_inador_route_get_dest (route, &dest, NULL);
	
	inet_ntop (family, &dest, buffer_ip, sizeof (buffer_ip));
	
	printf ("___ MANAGER ___ Informando ruta agregada: %s/%i [Tabla %i]\n", buffer_ip, prefix, table);
	
	for (g = manager->connected_client_list; g != NULL; g = g->next) {
		manager_client = (ManagerClientInfo *) g->data;
		
		if (manager_client->wanted_events & NET_INADOR_EVENT_MASK_ROUTES) {
			printf ("___ MANAGER ___ Informando a la conexión (%i)\n", manager_client->fd);
			_manager_send_route (manager_client, route, TRUE);
		}
	}
}

